"""Fetch data from PyPI."""

import asyncio
import hashlib
from typing import List, Optional

import aiohttp
from packaging.version import Version

from django.core.cache import cache

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from package_monitor import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

BASE_URL = "https://pypi.org/pypi"
CACHE_TIMEOUT = 3600 * 24
CACHE_KEY = "package-monitor-pypi-"


async def fetch_pypi_releases(
    session: aiohttp.ClientSession, name: str, releases: List[Version]
) -> List[dict]:
    """Fetch and return data for releases of a pypi project."""
    tasks = [
        asyncio.create_task(
            fetch_release_from_pypi_async(session, name=name, version=str(r))
        )
        for r in releases
    ]
    r = await asyncio.gather(*tasks)
    return r


async def fetch_project_from_pypi_async(
    session: aiohttp.ClientSession, name: str
) -> Optional[dict]:
    """Fetch project data from PyPI and return it.

    Returns None if there was an API error.
    """
    return await _fetch_data_from_pypi_async(session, _make_pypi_url(name))


async def fetch_release_from_pypi_async(
    session: aiohttp.ClientSession, name: str, version: str
) -> Optional[dict]:
    """Fetch release data from PyPI and return it.

    Returns None if there was an API error.
    """
    key = _make_cache_key(name, version)
    if data := await cache.aget(key):
        return data

    r = await _fetch_data_from_pypi_async(session, _make_pypi_url(name, version))
    await cache.aset(key=key, value=r, timeout=CACHE_TIMEOUT)
    return r


def _make_cache_key(name: str, version: str) -> str:
    b = f"{name}-{version}".encode("utf-8")
    key_hash = hashlib.md5(b).hexdigest()
    key = f"{CACHE_KEY}{key_hash}"
    return key


def _make_pypi_url(name: str, version: Optional[str] = None) -> str:
    if not version:
        return f"{BASE_URL}/{name}/json"
    return f"{BASE_URL}/{name}/{version}/json"


async def _fetch_data_from_pypi_async(
    session: aiohttp.ClientSession, url: str
) -> Optional[dict]:
    """Fetch JSON data for a URL and return it.

    Returns None if there was an API error.
    """
    logger.info("Fetching data from PyPI for url: %s", url)

    async with session.get(url) as resp:
        if not resp.ok:
            if resp.status == 404:
                logger.info("PyPI URL not found: %s", url)
            else:
                logger.warning(
                    "Failed to retrieve data from PyPI for "
                    "url '%s'. "
                    "Status code: %d, "
                    "response: %s",
                    url,
                    resp.status,
                    await resp.text(),
                )
            return None

        data = await resp.json()
        return data


def clear_cache():
    """Clear the release cache."""
    keys = cache.iter_keys(f"{CACHE_KEY}*")
    cache.delete_many(keys)



"""Core logic for parsed distribution packages."""

import asyncio
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional

import aiohttp
import importlib_metadata
from packaging.markers import UndefinedComparison, UndefinedEnvironmentName
from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import InvalidVersion, Version
from packaging.version import parse as version_parse

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from package_monitor import __title__
from package_monitor.app_settings import (
    PACKAGE_MONITOR_CUSTOM_REQUIREMENTS,
    PACKAGE_MONITOR_PROTECTED_PACKAGES,
)

from . import metadata_helpers
from .pypi import fetch_project_from_pypi_async, fetch_pypi_releases

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@dataclass
class DistributionPackage:
    """A parsed distribution package."""

    name: str
    current: str
    is_editable: bool
    requirements: List[Requirement] = field(default_factory=list)
    apps: List[str] = field(default_factory=list)
    latest: str = ""
    homepage_url: str = ""
    summary: str = ""

    def __str__(self) -> str:
        return f"{self.name} {self.current}"

    @property
    def name_normalized(self) -> str:
        """Return normalized name."""
        return canonicalize_name(self.name)

    def is_outdated(self) -> Optional[bool]:
        """Is this package outdated?"""
        if self.current and self.latest:
            return version_parse(self.current) < version_parse(self.latest)
        return None

    def is_prerelease(self) -> bool:
        """Determine if this package is a prerelease."""
        current_version = version_parse(self.current)
        current_is_prerelease = (
            str(current_version) == str(self.current) and current_version.is_prerelease
        )
        return current_is_prerelease

    def _package_specifiers_from_requirements(self, requirements: dict) -> SpecifierSet:
        """Return consolidated specifiers for this package from all other packages."""
        s = SpecifierSet()
        if self.name_normalized in requirements:
            for _, specifier in requirements[self.name_normalized].items():
                s &= specifier
        return s

    async def update_from_pypi_async(
        self,
        session: aiohttp.ClientSession,
        requirements: dict,
        protected_packages_versions: dict,
        system_python: Version,
    ) -> bool:
        """Update latest version and URL from PyPI.

        Return True if update was successful, else False.
        """

        pypi_data = await fetch_project_from_pypi_async(session, name=self.name)
        if not pypi_data:
            return False

        updates = self._determine_available_updates(
            pypi_data_releases=pypi_data["releases"],
            package_requirements=self._package_specifiers_from_requirements(
                requirements
            ),
            system_python=system_python,
        )
        latest = await self._determine_latest_available_update(
            session,
            updates=updates,
            protected_packages_versions=protected_packages_versions,
        )

        self.latest = str(latest) if latest else self.current

        pypi_info = pypi_data.get("info")
        pypi_url = pypi_info.get("project_url", "") if pypi_info else ""
        self.homepage_url = pypi_url
        return True

    def _determine_available_updates(
        self,
        pypi_data_releases: dict,
        package_requirements: SpecifierSet,
        system_python: Version,
    ) -> List[Version]:
        """Determine latest valid updates available on PyPI
        and return them as ascending list.
        """
        updates = []
        current_version = (
            version_parse(self.current) if self.current else Version("0.0.0")
        )
        for release, release_details in pypi_data_releases.items():
            version = self._release_version(release)
            if not version:
                continue

            if version.is_prerelease and not self.is_prerelease():
                continue

            if not self._release_is_valid(package_requirements, version):
                continue

            if version <= current_version:
                continue

            release_detail = release_details[-1] if len(release_details) > 0 else None
            if release_detail:
                if release_detail["yanked"]:
                    continue

                if not self._required_python_matches(release_detail, system_python):
                    continue

            updates.append(version)

        return updates

    def _release_version(self, version_string: str) -> Optional[Version]:
        try:
            version = version_parse(version_string)
        except InvalidVersion:
            logger.info(
                "%s: Ignoring release with invalid version: %s",
                self.name,
                version_string,
            )
            return None

        if str(version) != str(version_string):
            return None

        return version

    def _release_is_valid(
        self, package_requirements: SpecifierSet, version: Version
    ) -> bool:
        if len(package_requirements) == 0:
            return True

        return version in package_requirements

    def _required_python_matches(
        self, release_detail, system_python_version: Version
    ) -> bool:
        if requires_python := release_detail.get("requires_python"):
            try:
                required_python_versions = SpecifierSet(requires_python)
            except InvalidSpecifier:
                logger.info(
                    "%s: Ignoring release with invalid requires_python: %s",
                    self.name,
                    requires_python,
                )
                return False

            if system_python_version not in required_python_versions:
                return False

        return True

    async def _determine_latest_available_update(
        self,
        session: aiohttp.ClientSession,
        updates: List[Version],
        protected_packages_versions: Dict[str, Version],
    ) -> Optional[Version]:
        """Determines latest available and valid update and returns it.
        Or return None if none are available.
        """
        if not updates:
            return None

        if protected_packages_versions:
            valid_updates = await self._gather_valid_updates(
                session, updates, protected_packages_versions
            )
        else:
            valid_updates = updates

        valid_updates.sort()
        latest = valid_updates.pop() if valid_updates else None
        return latest

    async def _gather_valid_updates(self, session, updates, package_versions):
        valid_updates = []
        releases = await fetch_pypi_releases(session, name=self.name, releases=updates)
        for release in releases:
            info = release.get("info")
            if not info:
                continue

            found_issue = False
            for req_str in info.get("requires_dist", []):
                try:
                    r = Requirement(req_str)
                except InvalidRequirement:
                    continue  # invalid requirements can be ignored

                if not is_marker_valid(r):
                    continue  # invalid requirements can be ignored

                if (name := canonicalize_name(r.name)) in package_versions:
                    v = package_versions[name]
                    if v not in r.specifier:
                        logger.debug(
                            "%s: Update does not match current packages: %s", self, r
                        )
                        found_issue = True
                        break  # exit at first found issue

            if found_issue:
                continue

            update = version_parse(info["version"])
            valid_updates.append(update)

        return valid_updates

    @classmethod
    def create_from_metadata_distribution(
        cls, dist: importlib_metadata.Distribution, disable_app_check=False
    ):
        """Create new object from a metadata distribution.

        This is the only place where we are accessing the importlib metadata API
        for a specific distribution package and are thus storing
        all needed information about that package in our new object.
        Should additional information be needed sometimes it should be fetched here too.
        """
        obj = cls(
            name=dist.name,
            current=dist.version,
            is_editable=metadata_helpers.is_distribution_editable(dist),
            requirements=metadata_helpers.parse_requirements(dist),
            summary=metadata_helpers.metadata_value(dist, "Summary"),
        )
        if not disable_app_check:
            obj.apps = metadata_helpers.identify_installed_django_apps(dist)
        return obj


def gather_distribution_packages() -> Dict[str, DistributionPackage]:
    """Gather distribution packages and detect Django apps."""
    packages = [
        DistributionPackage.create_from_metadata_distribution(dist)
        for dist in importlib_metadata.distributions()
        if dist.metadata["Name"]
    ]
    return {obj.name_normalized: obj for obj in packages}


def compile_package_requirements(
    packages: Dict[str, DistributionPackage]
) -> Dict[str, Dict[str, SpecifierSet]]:
    """Consolidate requirements from all known distributions and known packages"""
    requirements = defaultdict(dict)

    # add requirements from all packages
    for package in packages.values():
        for requirement in package.requirements:
            _add_valid_requirement(requirements, requirement, package.name, packages)

    # add requirements from settings (if any)
    for requirement_string in PACKAGE_MONITOR_CUSTOM_REQUIREMENTS:
        try:
            requirement = Requirement(requirement_string)
        except InvalidRequirement:
            continue
        _add_valid_requirement(requirements, requirement, "CUSTOM", packages)

    return dict(requirements)


def _add_valid_requirement(
    requirements: dict, requirement: Requirement, package_name: str, packages: dict
):
    name = canonicalize_name(requirement.name)
    if name not in packages:
        return

    if not is_marker_valid(requirement):
        return

    requirements[name][package_name] = requirement.specifier


def is_marker_valid(requirement: Requirement) -> bool:
    """Report wether a requirement is valid based on it's marker.
    No marker means also True.
    """
    if not requirement.marker:
        return True

    try:
        return requirement.marker.evaluate()
    except (UndefinedEnvironmentName, UndefinedComparison):
        return False


def update_packages_from_pypi(
    packages: Dict[str, DistributionPackage], requirements: dict
) -> None:
    """Update packages with latest versions and URL from PyPI in accordance
    with the given requirements and updates the packages.
    """

    async def update_packages_from_pypi_async() -> None:
        """Update packages from PyPI concurrently."""
        system_python_version = determine_system_python_version()
        packages_versions = gather_protected_packages_versions(packages)
        async with aiohttp.ClientSession() as session:
            tasks = [
                asyncio.create_task(
                    package.update_from_pypi_async(
                        session=session,
                        requirements=requirements,
                        protected_packages_versions=packages_versions,
                        system_python=system_python_version,
                    )
                )
                for package in packages.values()
            ]
            await asyncio.gather(*tasks)

    asyncio.run(update_packages_from_pypi_async())


def determine_system_python_version() -> Version:
    """Return current Python version of this system."""
    result = version_parse(
        f"{sys.version_info.major}.{sys.version_info.minor}"
        f".{sys.version_info.micro}"
    )
    return result


def gather_protected_packages_versions(
    packages: Dict[str, DistributionPackage]
) -> Dict[str, Version]:
    """Return versions of protected packages
    or empty when no protected packages are defined or matching.
    """
    focus_packages = set(PACKAGE_MONITOR_PROTECTED_PACKAGES)
    if not focus_packages:
        return {}

    focus_names = {canonicalize_name(p) for p in focus_packages}
    result = {
        name: version_parse(p.current)
        for name, p in packages.items()
        if name in focus_names
    }
    return result



from app_utils.testing import NoSocketsTestCase

from .factories import DistributionFactory


class TestDistribution(NoSocketsTestCase):
    def test_should_update_has_installed_apps_1(self):
        # given
        obj = DistributionFactory()
        # when
        obj.calc_has_installed_apps()
        # then
        self.assertFalse(obj.has_installed_apps)

    def test_should_update_has_installed_apps_2(self):
        # given
        obj = DistributionFactory(apps=["dummy"])
        # when
        obj.calc_has_installed_apps()
        # then
        self.assertTrue(obj.has_installed_apps)

    def test_should_update_has_installed_apps_when_saved_1(self):
        # when
        obj = DistributionFactory()
        # then
        self.assertFalse(obj.has_installed_apps)

    def test_should_update_has_installed_apps_when_saved_2(self):
        # when
        obj = DistributionFactory(apps=["dummy"])
        # then
        self.assertTrue(obj.has_installed_apps)



from unittest import mock

from packaging.specifiers import SpecifierSet

from app_utils.testing import NoSocketsTestCase

from package_monitor.models import Distribution

from .factories import DistributionFactory, DistributionPackageFactory, make_packages

MODULE_PATH = "package_monitor.managers"


@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", False)
@mock.patch(MODULE_PATH + ".update_packages_from_pypi", spec=True)
@mock.patch(MODULE_PATH + ".compile_package_requirements", spec=True)
@mock.patch(MODULE_PATH + ".gather_distribution_packages", spec=True)
class TestDistributionsUpdateAll(NoSocketsTestCase):
    def test_should_create_new_packages_from_scratch(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="alpha",
            current="1.0.0",
            homepage_url="https://www.alpha.com",
            summary="alpha-description",
        )
        dist_bravo = DistributionPackageFactory(
            name="bravo",
            requires=["alpha>=1.0.0"],
            homepage_url="https://www.bravo.com",
        )
        packages = make_packages(dist_alpha, dist_bravo)
        packages["alpha"].apps = ["alpha_app"]
        mock_gather_distribution_packages.return_value = packages
        mock_compile_package_requirements.return_value = {
            "alpha": {"bravo": SpecifierSet(">=1.0.0")}
        }
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 2)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)
        self.assertTrue(obj.has_installed_apps)
        self.assertEqual(obj.website_url, "https://www.alpha.com")
        self.assertEqual(obj.description, "alpha-description")
        self.assertListEqual(obj.apps, ["alpha_app"])
        self.assertEqual(
            obj.used_by,
            [
                {
                    "name": "bravo",
                    "homepage_url": "https://www.bravo.com",
                    "requirements": [">=1.0.0"],
                }
            ],
        )

    def test_should_retain_package_name_with_capitals(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="Alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.first()
        self.assertEqual(obj.name, "Alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_update_existing_packages(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_remove_stale_packages(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        DistributionFactory(name="bravo", installed_version="1.0.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_set_is_outdated_to_none_when_no_pypi_infos(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.latest_version, "")
        self.assertIsNone(obj.is_outdated)

    def test_should_set_is_outdated_to_none_when_current_version_can_not_be_parsed(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="2009r")
        packages = make_packages(dist_alpha)
        packages["alpha"].latest = ""
        mock_gather_distribution_packages.return_value = packages
        mock_compile_package_requirements.return_value = {}
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.latest_version, "")
        self.assertIsNone(obj.is_outdated)


@mock.patch(MODULE_PATH + ".notify_admins", spec=True)
@mock.patch(MODULE_PATH + ".update_packages_from_pypi", spec=True)
@mock.patch(MODULE_PATH + ".compile_package_requirements", spec=True)
@mock.patch(MODULE_PATH + ".gather_distribution_packages", spec=True)
class TestDistributionsUpdateAllNotifications(NoSocketsTestCase):
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True)
    def test_should_notify_when_enabled_and_there_is_an_update(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0"
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True):
            Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "1.1.0")
        self.assertTrue(mock_notify_admins.called)

    def test_should_not_notify_when_disabled_via_setting(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0"
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", False):
            Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "")
        self.assertFalse(mock_notify_admins.called)

    def test_should_notify_when_outdated_and_newer_update(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0"
        )
        DistributionFactory(
            name="Alpha",
            installed_version="1.0.0",
            latest_version="1.1.0",
            is_outdated=True,
            latest_notified_version="1.0.0",
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True):
            Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "1.1.0")
        self.assertTrue(mock_notify_admins.called)

    def test_should_not_notify_when_already_notified(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0"
        )
        DistributionFactory(
            name="Alpha",
            installed_version="1.0.0",
            latest_version="1.1.0",
            is_outdated=True,
            latest_notified_version="1.1.0",
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True):
            Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "1.1.0")
        self.assertFalse(mock_notify_admins.called)

    def test_should_not_notify_when_disabled_directly(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0"
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True):
            Distribution.objects.update_all(notifications_disabled=True)
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "")
        self.assertFalse(mock_notify_admins.called)

    def test_should_not_notify_when_editable_and_those_are_not_shown(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
        mock_notify_admins,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="Alpha", current="1.0.0", latest="1.1.0", is_editable=True
        )
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        with mock.patch(
            MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True
        ), mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False):
            Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="Alpha")
        self.assertTrue(obj.is_outdated)
        self.assertEqual(obj.latest_notified_version, "")
        self.assertFalse(mock_notify_admins.called)


class TestDistributionFilterVisible(NoSocketsTestCase):
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_all_packages(self):
        # given
        obj_1 = DistributionFactory()
        obj_2 = DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name, obj_2.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_apps_only(self):
        # given
        obj_1 = DistributionFactory(apps=["app_1"])
        DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", ["include-me"])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_apps_plus_included(self):
        # given
        obj_1 = DistributionFactory(apps=["app_1"])
        obj_2 = DistributionFactory(name="include-me")
        DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name, obj_2.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", ["exclude-me"])
    def test_should_have_all_packages_minus_excluded(self):
        # given
        obj_1 = DistributionFactory()
        DistributionFactory(name="exclude-me")
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_all_packages_minus_editable(self):
        # given
        obj_1 = DistributionFactory()
        DistributionFactory(name="exclude-me", is_editable=True)
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})


@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
class TestDistributionBuildInstallCommand(NoSocketsTestCase):
    def test_all_packages(self):
        # given
        DistributionFactory(name="alpha", latest_version="1.2.0")
        DistributionFactory(name="bravo", latest_version="2.1.0")
        # when
        result = Distribution.objects.order_by("name").build_install_command()
        # then
        self.assertEqual(result, "pip install alpha==1.2.0 bravo==2.1.0")

    def test_should_stay_within_max_line_length(self):
        # given
        DistributionFactory.create_batch(size=500)
        # when
        result = Distribution.objects.all().build_install_command()
        # then
        print(result)
        print(len(result))
        self.assertLessEqual(len(result), 4095)



"""An app that helps keep track of installed packages
and outstanding updates for Alliance Auth.
"""

# pylint: disable=invalid-name
default_app_config = "package_monitor.apps.PackageMonitorConfig"

__version__ = "1.15.0"
__title__ = "Package Monitor"



from unittest import IsolatedAsyncioTestCase

import aiohttp
from aioresponses import aioresponses
from packaging.version import Version

from package_monitor.core.pypi import (
    clear_cache,
    fetch_project_from_pypi_async,
    fetch_pypi_releases,
    fetch_release_from_pypi_async,
)
from package_monitor.tests.factories import DistributionPackageFactory, PypiFactory


class TestFetchDataFromPypi(IsolatedAsyncioTestCase):
    @aioresponses()
    async def test_should_return_data(self, requests_mocker: aioresponses):
        # given
        requests_mocker.get("https://pypi.org/pypi/alpha/json", payload={"alpha": 1})
        # when
        async with aiohttp.ClientSession() as session:
            result = await fetch_project_from_pypi_async(session, "alpha")
        # then
        self.assertEqual(result, {"alpha": 1})

    @aioresponses()
    async def test_should_return_none_when_package_does_not_exist(
        self, requests_mocker: aioresponses
    ):
        # given
        requests_mocker.get("https://pypi.org/pypi/alpha/json", status=404)
        # when
        async with aiohttp.ClientSession() as session:
            result = await fetch_project_from_pypi_async(session, "alpha")
        # then
        self.assertIsNone(result)

    @aioresponses()
    async def test_should_return_none_on_other_http_errors(
        self, requests_mocker: aioresponses
    ):
        # given
        requests_mocker.get("https://pypi.org/pypi/alpha/json", status=500)
        # when
        async with aiohttp.ClientSession() as session:
            result = await fetch_project_from_pypi_async(session, "alpha")
        # then
        self.assertIsNone(result)


class TestFetchPypiReleases(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        clear_cache()

    @aioresponses()
    async def test_should_return_data(self, requests_mocker: aioresponses):
        # given
        dist_1 = DistributionPackageFactory(name="alpha", current="1.2.3")
        pypi_1 = PypiFactory(distribution=dist_1)
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/1.2.3/json", payload=pypi_1.asdict()
        )
        dist_2 = DistributionPackageFactory(name="alpha", current="1.2.5")
        pypi_2 = PypiFactory(distribution=dist_2)
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/1.2.5/json", payload=pypi_2.asdict()
        )
        versions = [Version("1.2.3"), Version("1.2.5")]

        # when
        async with aiohttp.ClientSession() as session:
            result = await fetch_pypi_releases(session, "alpha", versions)

        # then
        self.assertEqual(len(result), 2)
        p = result[0]
        self.assertEqual(p["info"]["name"], "alpha")
        self.assertEqual(p["info"]["version"], "1.2.3")
        p = result[1]
        self.assertEqual(p["info"]["name"], "alpha")
        self.assertEqual(p["info"]["version"], "1.2.5")


class TestFetchPypiRelease(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        clear_cache()

    @aioresponses()
    async def test_should_return_data(self, requests_mocker: aioresponses):
        # given
        dist = DistributionPackageFactory(name="alpha", current="1.2.3")
        pypi = PypiFactory(distribution=dist)
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/1.2.3/json", payload=pypi.asdict()
        )

        # when
        async with aiohttp.ClientSession() as session:
            result = await fetch_release_from_pypi_async(session, "alpha", "1.2.3")

        # then
        self.assertEqual(result["info"]["name"], "alpha")
        self.assertEqual(result["info"]["version"], "1.2.3")
        requests_mocker.assert_called_once()

    @aioresponses()
    async def test_should_use_cache(self, requests_mocker: aioresponses):
        # given
        dist = DistributionPackageFactory(name="alpha", current="1.2.3")
        pypi = PypiFactory(distribution=dist)
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/1.2.3/json", payload=pypi.asdict()
        )

        # when
        async with aiohttp.ClientSession() as session:
            await fetch_release_from_pypi_async(session, "alpha", "1.2.3")
            await fetch_release_from_pypi_async(session, "alpha", "1.2.3")
            result = await fetch_release_from_pypi_async(session, "alpha", "1.2.3")

        # then
        self.assertEqual(result["info"]["version"], "1.2.3")
        requests_mocker.assert_called_once()



from unittest import TestCase, mock

from aioresponses import aioresponses

from package_monitor import tasks
from package_monitor.core import pypi
from package_monitor.core.distribution_packages import DistributionPackage
from package_monitor.models import Distribution

from .factories import MetadataDistributionStubFactory, PypiFactory, PypiReleaseFactory

CORE_PATH = "package_monitor.core.distribution_packages"
CORE_HELPERS_PATH = "package_monitor.core.metadata_helpers"
MANAGERS_PATH = "package_monitor.managers"


@mock.patch(MANAGERS_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", False)
@mock.patch(CORE_HELPERS_PATH + ".django_apps", spec=True)
@mock.patch(CORE_PATH + ".importlib_metadata.distributions", spec=True)
class TestUpdatePackagesFromPyPi(TestCase):
    @classmethod
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        # workaround to remove obj which is not cleaned up
        Distribution.objects.all().delete()

    def setUp(self) -> None:
        pypi.clear_cache()

    @aioresponses()
    def test_should_update_packages(
        self, mock_distributions, mock_django_apps, requests_mocker
    ):
        # given
        dist_alpha = MetadataDistributionStubFactory(name="alpha", version="1.0.0")
        distributions = lambda: iter([dist_alpha])  # noqa: E731
        mock_distributions.side_effect = distributions
        mock_django_apps.get_app_configs.return_value = []

        pypi_alpha = PypiFactory(
            distribution=DistributionPackage.create_from_metadata_distribution(
                dist_alpha
            )
        )
        pypi_alpha.info.version = "1.1.0"
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/json", payload=pypi_alpha.asdict()
        )
        requests_mocker.get(
            "https://pypi.org/pypi/alpha/1.1.0/json", payload=pypi_alpha.asdict()
        )

        # when
        tasks.update_distributions()

        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.latest_version, "1.1.0")
